@@ -7,7 +7,9 @@
 
 #pragma once
 
+#include <QObject>
 #include <QSqlQuery>
+#include <QString>
 
 class Database : public QObject
 {
@@ -22,6 +24,9 @@ public:
     bool execute(QSqlQuery &query);
     bool execute(const QString &query);
 
+    bool transaction();
+    bool commit();
+
 private:
     Database();
     int version();
@@ -31,5 +36,7 @@ private:
     bool migrateTo2();
     bool migrateTo3();
     bool migrateTo4();
+    bool migrateTo5();
+    bool migrateTo6();
     void cleanup();
 };
