@@ -5,6 +5,7 @@
  */
 
 #include "datamanager.h"
+#include "datamanagerlogging.h"
 
 #include <QDateTime>
 #include <QDir>
@@ -17,12 +18,12 @@
 
 #include "audiomanager.h"
 #include "database.h"
-#include "datamanagerlogging.h"
 #include "entry.h"
 #include "feed.h"
 #include "fetcher.h"
 #include "settingsmanager.h"
 #include "storagemanager.h"
+#include "sync/sync.h"
 
 DataManager::DataManager()
 {
@@ -49,11 +50,11 @@ DataManager::DataManager()
             }
         });
     connect(&Fetcher::instance(), &Fetcher::entryAdded, this, [this](const QString &feedurl, const QString &id) {
+        Q_UNUSED(feedurl)
         // Only add the new entry to m_entries
         // we will repopulate m_entrymap once all new entries have been added,
         // such that m_entrymap will show all new entries in the correct order
         m_entries[id] = nullptr;
-        Q_EMIT entryAdded(feedurl, id);
     });
     connect(&Fetcher::instance(), &Fetcher::feedUpdated, this, [this](const QString &feedurl) {
         // Update m_entrymap for feedurl, such that the new and old entries show
@@ -76,7 +77,7 @@ DataManager::DataManager()
             Database::instance().execute(query);
             while (query.next()) {
                 QString id = query.value(QStringLiteral("id")).toString();
-                addToQueue(feedurl, id);
+                getEntry(id)->setQueueStatusInternal(true);
                 if (SettingsManager::self()->autoDownload()) {
                     if (getEntry(id) && getEntry(id)->hasEnclosure() && getEntry(id)->enclosure()) {
                         qCDebug(kastsDataManager) << "Start downloading" << getEntry(id)->title();
@@ -147,56 +148,16 @@ Entry *DataManager::getEntry(const QString &id) const
     return m_entries[id];
 }
 
-Entry *DataManager::getEntry(const EpisodeModel::Type type, const int entry_index) const
-{
-    QSqlQuery entryQuery;
-    if (type == EpisodeModel::All || type == EpisodeModel::New || type == EpisodeModel::Unread || type == EpisodeModel::Downloading
-        || type == EpisodeModel::PartiallyDownloaded || type == EpisodeModel::Downloaded) {
-        if (type == EpisodeModel::New) {
-            entryQuery.prepare(QStringLiteral("SELECT id FROM Entries WHERE new=:new ORDER BY updated DESC LIMIT 1 OFFSET :index;"));
-            entryQuery.bindValue(QStringLiteral(":new"), true);
-        } else if (type == EpisodeModel::Unread) {
-            entryQuery.prepare(QStringLiteral("SELECT id FROM Entries WHERE read=:read ORDER BY updated DESC LIMIT 1 OFFSET :index;"));
-            entryQuery.bindValue(QStringLiteral(":read"), false);
-        } else if (type == EpisodeModel::All) {
-            entryQuery.prepare(QStringLiteral("SELECT id FROM Entries ORDER BY updated DESC LIMIT 1 OFFSET :index;"));
-        } else if (type == EpisodeModel::Downloading) {
-            entryQuery.prepare(
-                QStringLiteral("SELECT * FROM Enclosures INNER JOIN Entries ON Enclosures.id = Entries.id WHERE downloaded=:downloaded ORDER BY updated DESC "
-                               "LIMIT 1 OFFSET :index;"));
-            entryQuery.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloading));
-        } else if (type == EpisodeModel::PartiallyDownloaded) {
-            entryQuery.prepare(
-                QStringLiteral("SELECT * FROM Enclosures INNER JOIN Entries ON Enclosures.id = Entries.id WHERE downloaded=:downloaded ORDER BY updated DESC "
-                               "LIMIT 1 OFFSET :index;"));
-            entryQuery.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::PartiallyDownloaded));
-        } else if (type == EpisodeModel::Downloaded) {
-            entryQuery.prepare(
-                QStringLiteral("SELECT * FROM Enclosures INNER JOIN Entries ON Enclosures.id = Entries.id WHERE downloaded=:downloaded ORDER BY updated DESC "
-                               "LIMIT 1 OFFSET :index;"));
-            entryQuery.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloaded));
-        } else {
-            // this should not happen
-            qWarning() << "Cannot find entry type" << type << "in getEntry for entry index" << entry_index;
-        }
-        entryQuery.bindValue(QStringLiteral(":index"), entry_index);
-        Database::instance().execute(entryQuery);
-        if (!entryQuery.next()) {
-            qWarning() << "No element with index" << entry_index << "found";
-            return nullptr;
-        }
-        QString id = entryQuery.value(QStringLiteral("id")).toString();
-        return getEntry(id);
-    }
-    qWarning() << "Cannot find entry type" << type << "in getEntry for entry index" << entry_index;
-    return nullptr;
-}
-
 int DataManager::feedCount() const
 {
     return m_feedmap.count();
 }
 
+QStringList DataManager::getIdList(const Feed *feed) const
+{
+    return m_entrymap[feed->url()];
+}
+
 int DataManager::entryCount(const int feed_index) const
 {
     return m_entrymap[m_feedmap[feed_index]].count();
@@ -207,41 +168,6 @@ int DataManager::entryCount(const Feed *feed) const
     return m_entrymap[feed->url()].count();
 }
 
-int DataManager::entryCount(const EpisodeModel::Type type) const
-{
-    QSqlQuery query;
-    if (type == EpisodeModel::All || type == EpisodeModel::New || type == EpisodeModel::Unread || type == EpisodeModel::Downloading
-        || type == EpisodeModel::PartiallyDownloaded || type == EpisodeModel::Downloaded) {
-        if (type == EpisodeModel::New) {
-            query.prepare(QStringLiteral("SELECT COUNT (id) FROM Entries WHERE new=:new;"));
-            query.bindValue(QStringLiteral(":new"), true);
-        } else if (type == EpisodeModel::Unread) {
-            query.prepare(QStringLiteral("SELECT COUNT (id) FROM Entries WHERE read=:read;"));
-            query.bindValue(QStringLiteral(":read"), false);
-        } else if (type == EpisodeModel::All) {
-            query.prepare(QStringLiteral("SELECT COUNT (id) FROM Entries;"));
-        } else if (type == EpisodeModel::Downloading) {
-            query.prepare(QStringLiteral("SELECT COUNT (id) FROM Enclosures WHERE downloaded=:downloaded;"));
-            query.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloading));
-        } else if (type == EpisodeModel::PartiallyDownloaded) {
-            query.prepare(QStringLiteral("SELECT COUNT (id) FROM Enclosures WHERE downloaded=:downloaded;"));
-            query.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::PartiallyDownloaded));
-        } else if (type == EpisodeModel::Downloaded) {
-            query.prepare(QStringLiteral("SELECT COUNT (id) FROM Enclosures WHERE downloaded=:downloaded;"));
-            query.bindValue(QStringLiteral(":downloaded"), Enclosure::statusToDb(Enclosure::Downloaded));
-        } else {
-            // this should not happen
-            qWarning() << "Cannot find entry type" << type << "in entryCount";
-        }
-        Database::instance().execute(query);
-        if (!query.next())
-            return -1;
-        return query.value(0).toInt();
-    }
-    qWarning() << "Cannot find entry type" << type << "in entryCount";
-    return -1;
-}
-
 int DataManager::unreadEntryCount(const Feed *feed) const
 {
     QSqlQuery query;
@@ -266,72 +192,122 @@ int DataManager::newEntryCount(const Feed *feed) const
 
 void DataManager::removeFeed(Feed *feed)
 {
-    qCDebug(kastsDataManager) << "deleting feed" << feed->url() << "with index" << m_feedmap.indexOf(feed->url());
-    removeFeed(m_feedmap.indexOf(feed->url()));
+    QList<Feed *> feeds;
+    feeds << feed;
+    removeFeeds(feeds);
 }
 
 void DataManager::removeFeed(const int index)
 {
     // Get feed pointer
     Feed *feed = getFeed(m_feedmap[index]);
-    const QString feedurl = feed->url();
+    removeFeed(feed);
+}
 
-    // Delete the object instances and mappings
-    // First delete entries in Queue
-    qCDebug(kastsDataManager) << "delete queueentries of" << feedurl;
-    for (auto &id : m_queuemap) {
-        if (getEntry(id)->feed()->url() == feedurl) {
-            if (AudioManager::instance().entry() == getEntry(id)) {
-                AudioManager::instance().next();
-            }
-            removeQueueItem(id);
+void DataManager::removeFeeds(const QStringList &feedurls)
+{
+    QList<Feed *> feeds;
+    for (QString feedurl : feedurls) {
+        feeds << getFeed(feedurl);
+    }
+    removeFeeds(feeds);
+}
+
+void DataManager::removeFeeds(const QVariantList feedVariantList)
+{
+    QList<Feed *> feeds;
+    for (QVariant feedVariant : feedVariantList) {
+        if (feedVariant.canConvert<Feed *>()) {
+            feeds << feedVariant.value<Feed *>();
         }
     }
+    removeFeeds(feeds);
+}
 
-    // Delete entries themselves
-    qCDebug(kastsDataManager) << "delete entries of" << feedurl;
-    for (auto &id : m_entrymap[feedurl]) {
-        if (getEntry(id)->hasEnclosure())
-            getEntry(id)->enclosure()->deleteFile(); // delete enclosure (if it exists)
-        if (!getEntry(id)->image().isEmpty())
-            StorageManager::instance().removeImage(getEntry(id)->image()); // delete entry images
-        delete m_entries[id]; // delete pointer
-        m_entries.remove(id); // delete the hash key
+void DataManager::removeFeeds(const QList<Feed *> &feeds)
+{
+    for (Feed *feed : feeds) {
+        const QString feedurl = feed->url();
+        int index = m_feedmap.indexOf(feedurl);
+
+        qCDebug(kastsDataManager) << "deleting feed" << feedurl << "with index" << index;
+
+        // Delete the object instances and mappings
+        // First delete entries in Queue
+        qCDebug(kastsDataManager) << "delete queueentries of" << feedurl;
+        QStringList removeFromQueueList;
+        for (auto &id : m_queuemap) {
+            if (getEntry(id)->feed()->url() == feedurl) {
+                if (AudioManager::instance().entry() == getEntry(id)) {
+                    AudioManager::instance().next();
+                }
+                removeFromQueueList += id;
+            }
+        }
+        bulkQueueStatus(false, removeFromQueueList);
+
+        // Delete entries themselves
+        qCDebug(kastsDataManager) << "delete entries of" << feedurl;
+        for (auto &id : m_entrymap[feedurl]) {
+            if (getEntry(id)->hasEnclosure())
+                getEntry(id)->enclosure()->deleteFile(); // delete enclosure (if it exists)
+            if (!getEntry(id)->image().isEmpty())
+                StorageManager::instance().removeImage(getEntry(id)->image()); // delete entry images
+            delete m_entries[id]; // delete pointer
+            m_entries.remove(id); // delete the hash key
+        }
+        m_entrymap.remove(feedurl); // remove all the entry mappings belonging to the feed
+
+        qCDebug(kastsDataManager) << "Remove feed image" << feed->image() << "for feed" << feedurl;
+        if (!feed->image().isEmpty())
+            StorageManager::instance().removeImage(feed->image());
+        m_feeds.remove(m_feedmap[index]); // remove from m_feeds
+        m_feedmap.removeAt(index); // remove from m_feedmap
+        delete feed; // remove the pointer
+
+        // Then delete everything from the database
+        qCDebug(kastsDataManager) << "delete database part of" << feedurl;
+
+        // Delete related Errors
+        QSqlQuery query;
+        query.prepare(QStringLiteral("DELETE FROM Errors WHERE url=:url;"));
+        query.bindValue(QStringLiteral(":url"), feedurl);
+        Database::instance().execute(query);
+
+        // Delete Authors
+        query.prepare(QStringLiteral("DELETE FROM Authors WHERE feed=:feed;"));
+        query.bindValue(QStringLiteral(":feed"), feedurl);
+        Database::instance().execute(query);
+
+        // Delete Chapters
+        query.prepare(QStringLiteral("DELETE FROM Chapters WHERE feed=:feed;"));
+        query.bindValue(QStringLiteral(":feed"), feedurl);
+        Database::instance().execute(query);
+
+        // Delete Entries
+        query.prepare(QStringLiteral("DELETE FROM Entries WHERE feed=:feed;"));
+        query.bindValue(QStringLiteral(":feed"), feedurl);
+        Database::instance().execute(query);
+
+        // Delete Enclosures
+        query.prepare(QStringLiteral("DELETE FROM Enclosures WHERE feed=:feed;"));
+        query.bindValue(QStringLiteral(":feed"), feedurl);
+        Database::instance().execute(query);
+
+        // Delete Feed
+        query.prepare(QStringLiteral("DELETE FROM Feeds WHERE url=:url;"));
+        query.bindValue(QStringLiteral(":url"), feedurl);
+        Database::instance().execute(query);
+
+        // Save this action to the database (including timestamp) in order to be
+        // able to sync with remote services
+        Sync::instance().storeRemoveFeedAction(feedurl);
+
+        Q_EMIT feedRemoved(index);
     }
-    m_entrymap.remove(feedurl); // remove all the entry mappings belonging to the feed
 
-    qCDebug(kastsDataManager) << "Remove feed image" << feed->image() << "for feed" << feedurl;
-    if (!feed->image().isEmpty())
-        StorageManager::instance().removeImage(feed->image());
-    m_feeds.remove(m_feedmap[index]); // remove from m_feeds
-    m_feedmap.removeAt(index); // remove from m_feedmap
-    delete feed; // remove the pointer
-
-    // Then delete everything from the database
-    qCDebug(kastsDataManager) << "delete database part of" << feedurl;
-
-    // Delete Authors
-    QSqlQuery query;
-    query.prepare(QStringLiteral("DELETE FROM Authors WHERE feed=:feed;"));
-    query.bindValue(QStringLiteral(":feed"), feedurl);
-    Database::instance().execute(query);
-
-    // Delete Entries
-    query.prepare(QStringLiteral("DELETE FROM Entries WHERE feed=:feed;"));
-    query.bindValue(QStringLiteral(":feed"), feedurl);
-    Database::instance().execute(query);
-
-    // Delete Enclosures
-    query.prepare(QStringLiteral("DELETE FROM Enclosures WHERE feed=:feed;"));
-    query.bindValue(QStringLiteral(":feed"), feedurl);
-    Database::instance().execute(query);
-
-    // Delete Feed
-    query.prepare(QStringLiteral("DELETE FROM Feeds WHERE url=:url;"));
-    query.bindValue(QStringLiteral(":url"), feedurl);
-    Database::instance().execute(query);
-
-    Q_EMIT feedRemoved(index);
+    // if settings allow, then upload these changes immediately to sync server
+    Sync::instance().doQuickSync();
 }
 
 void DataManager::addFeed(const QString &url)
@@ -341,52 +317,65 @@ void DataManager::addFeed(const QString &url)
 
 void DataManager::addFeed(const QString &url, const bool fetch)
 {
-    // This method will add the relevant internal data structures, and then add
-    // a preliminary entry into the database.  Those details (as well as entries,
-    // authors and enclosures) will be updated by calling Fetcher::fetch() which
-    // will trigger a full update of the feed and all related items.
-    qCDebug(kastsDataManager) << "Adding feed";
-    if (feedExists(url)) {
-        qCDebug(kastsDataManager) << "Feed already exists";
-        return;
-    }
-    qCDebug(kastsDataManager) << "Feed does not yet exist";
-
-    QUrl urlFromInput = QUrl::fromUserInput(url);
-    QSqlQuery query;
-    query.prepare(QStringLiteral(
-        "INSERT INTO Feeds VALUES (:name, :url, :image, :link, :description, :deleteAfterCount, :deleteAfterType, :subscribed, :lastUpdated, :new, :notify);"));
-    query.bindValue(QStringLiteral(":name"), urlFromInput.toString());
-    query.bindValue(QStringLiteral(":url"), urlFromInput.toString());
-    query.bindValue(QStringLiteral(":image"), QLatin1String(""));
-    query.bindValue(QStringLiteral(":link"), QLatin1String(""));
-    query.bindValue(QStringLiteral(":description"), QLatin1String(""));
-    query.bindValue(QStringLiteral(":deleteAfterCount"), 0);
-    query.bindValue(QStringLiteral(":deleteAfterType"), 0);
-    query.bindValue(QStringLiteral(":subscribed"), QDateTime::currentDateTime().toSecsSinceEpoch());
-    query.bindValue(QStringLiteral(":lastUpdated"), 0);
-    query.bindValue(QStringLiteral(":new"), true);
-    query.bindValue(QStringLiteral(":notify"), false);
-    Database::instance().execute(query);
-
-    m_feeds[urlFromInput.toString()] = nullptr;
-    m_feedmap.append(urlFromInput.toString());
-
-    Q_EMIT feedAdded(urlFromInput.toString());
-
-    if (fetch)
-        Fetcher::instance().fetch(urlFromInput.toString());
+    addFeeds(QStringList(url), fetch);
 }
 
 void DataManager::addFeeds(const QStringList &urls)
+{
+    addFeeds(urls, true);
+}
+
+void DataManager::addFeeds(const QStringList &urls, const bool fetch)
 {
     if (urls.count() == 0)
         return;
 
-    for (int i = 0; i < urls.count(); i++) {
-        addFeed(urls[i], false); // add preliminary feed entries, but do not fetch yet
+    // This method will add the relevant internal data structures, and then add
+    // a preliminary entry into the database.  Those details (as well as entries,
+    // authors and enclosures) will be updated by calling Fetcher::fetch() which
+    // will trigger a full update of the feed and all related items.
+    for (QString url : urls) {
+        qCDebug(kastsDataManager) << "Adding feed";
+        if (feedExists(url)) {
+            qCDebug(kastsDataManager) << "Feed already exists";
+            continue;
+        }
+        qCDebug(kastsDataManager) << "Feed does not yet exist";
+
+        QUrl urlFromInput = QUrl::fromUserInput(url);
+        QSqlQuery query;
+        query.prepare(
+            QStringLiteral("INSERT INTO Feeds VALUES (:name, :url, :image, :link, :description, :deleteAfterCount, :deleteAfterType, :subscribed, "
+                           ":lastUpdated, :new, :notify);"));
+        query.bindValue(QStringLiteral(":name"), urlFromInput.toString());
+        query.bindValue(QStringLiteral(":url"), urlFromInput.toString());
+        query.bindValue(QStringLiteral(":image"), QLatin1String(""));
+        query.bindValue(QStringLiteral(":link"), QLatin1String(""));
+        query.bindValue(QStringLiteral(":description"), QLatin1String(""));
+        query.bindValue(QStringLiteral(":deleteAfterCount"), 0);
+        query.bindValue(QStringLiteral(":deleteAfterType"), 0);
+        query.bindValue(QStringLiteral(":subscribed"), QDateTime::currentDateTime().toSecsSinceEpoch());
+        query.bindValue(QStringLiteral(":lastUpdated"), 0);
+        query.bindValue(QStringLiteral(":new"), true);
+        query.bindValue(QStringLiteral(":notify"), false);
+        Database::instance().execute(query);
+
+        m_feeds[urlFromInput.toString()] = nullptr;
+        m_feedmap.append(urlFromInput.toString());
+
+        // Save this action to the database (including timestamp) in order to be
+        // able to sync with remote services
+        Sync::instance().storeAddFeedAction(urlFromInput.toString());
+
+        Q_EMIT feedAdded(urlFromInput.toString());
     }
-    Fetcher::instance().fetch(urls);
+
+    if (fetch) {
+        Fetcher::instance().fetch(urls);
+    }
+
+    // if settings allow, upload these changes immediately to sync servers
+    Sync::instance().doQuickSync();
 }
 
 Entry *DataManager::getQueueEntry(int index) const
@@ -406,23 +395,27 @@ QStringList DataManager::queue() const
 
 bool DataManager::entryInQueue(const Entry *entry)
 {
-    return entryInQueue(entry->feed()->url(), entry->id());
+    return entryInQueue(entry->id());
 }
 
-bool DataManager::entryInQueue(const QString &feedurl, const QString &id) const
+bool DataManager::entryInQueue(const QString &id) const
 {
-    Q_UNUSED(feedurl);
     return m_queuemap.contains(id);
 }
 
-void DataManager::addToQueue(const Entry *entry)
+void DataManager::moveQueueItem(const int from, const int to)
 {
-    if (entry != nullptr) {
-        return addToQueue(entry->feed()->url(), entry->id());
-    }
+    // First move the items in the internal data structure
+    m_queuemap.move(from, to);
+
+    // Then make sure that the database Queue table reflects these changes
+    updateQueueListnrs();
+
+    // Make sure that the QueueModel is aware of the changes so it can update
+    Q_EMIT queueEntryMoved(from, to);
 }
 
-void DataManager::addToQueue(const QString &feedurl, const QString &id)
+void DataManager::addToQueue(const QString &id)
 {
     // If item is already in queue, then stop here
     if (m_queuemap.contains(id))
@@ -439,43 +432,28 @@ void DataManager::addToQueue(const QString &feedurl, const QString &id)
     QSqlQuery query;
     query.prepare(QStringLiteral("INSERT INTO Queue VALUES (:index, :feedurl, :id, :playing);"));
     query.bindValue(QStringLiteral(":index"), index);
-    query.bindValue(QStringLiteral(":feedurl"), feedurl);
+    query.bindValue(QStringLiteral(":feedurl"), getEntry(id)->feed()->url());
     query.bindValue(QStringLiteral(":id"), id);
     query.bindValue(QStringLiteral(":playing"), false);
     Database::instance().execute(query);
 
-    // Set status to unplayed/unread when adding item to the queue
-    if (getEntry(id)) {
-        getEntry(id)->setRead(false);
-    }
-
     // Make sure that the QueueModel is aware of the changes
     Q_EMIT queueEntryAdded(index, id);
 }
 
-void DataManager::moveQueueItem(const int from, const int to)
+void DataManager::removeFromQueue(const QString &id)
 {
-    // First move the items in the internal data structure
-    m_queuemap.move(from, to);
+    if (!entryInQueue(id)) {
+        return;
+    }
 
-    // Then make sure that the database Queue table reflects these changes
-    updateQueueListnrs();
-
-    // Make sure that the QueueModel is aware of the changes so it can update
-    Q_EMIT queueEntryMoved(from, to);
-}
-
-void DataManager::removeQueueItem(const int index)
-{
+    const int index = m_queuemap.indexOf(id);
     qCDebug(kastsDataManager) << "Queuemap is now:" << m_queuemap;
-    // Unset "new" state
-    getEntry(m_queuemap[index])->setNew(false);
+    qCDebug(kastsDataManager) << "Queue index of item to be removed" << index;
 
-    const QString id = m_queuemap[index];
-
-    // Unload track from AudioManager if it's currently playing
+    // Move to next track if it's currently playing
     if (AudioManager::instance().entry() == getEntry(id)) {
-        AudioManager::instance().setEntry(nullptr);
+        AudioManager::instance().next();
     }
 
     // Remove the item from the internal data structure
@@ -483,26 +461,14 @@ void DataManager::removeQueueItem(const int index)
 
     // Then make sure that the database Queue table reflects these changes
     QSqlQuery query;
-    query.prepare(QStringLiteral("DELETE FROM Queue WHERE listnr=:listnr;"));
-    query.bindValue(QStringLiteral(":listnr"), index);
+    query.prepare(QStringLiteral("DELETE FROM Queue WHERE id=:id;"));
+    query.bindValue(QStringLiteral(":id"), id);
     Database::instance().execute(query);
-    // ... and update all other listnrs in Queue table
-    updateQueueListnrs();
 
     // Make sure that the QueueModel is aware of the change so it can update
     Q_EMIT queueEntryRemoved(index, id);
 }
 
-void DataManager::removeQueueItem(const QString id)
-{
-    removeQueueItem(m_queuemap.indexOf(id));
-}
-
-void DataManager::removeQueueItem(Entry *entry)
-{
-    removeQueueItem(m_queuemap.indexOf(entry->id()));
-}
-
 QString DataManager::lastPlayingEntry()
 {
     QSqlQuery query;
@@ -591,9 +557,17 @@ void DataManager::exportFeeds(const QString &path)
     xmlWriter.writeEndDocument();
 }
 
-void DataManager::loadFeed(const QString feedurl) const
+void DataManager::loadFeed(const QString &feedurl) const
 {
-    m_feeds[feedurl] = new Feed(feedurl);
+    QSqlQuery query;
+    query.prepare(QStringLiteral("SELECT url FROM Feeds WHERE url=:feedurl;"));
+    query.bindValue(QStringLiteral(":feedurl"), feedurl);
+    Database::instance().execute(query);
+    if (!query.next()) {
+        qWarning() << "Failed to load feed" << feedurl;
+    } else {
+        m_feeds[feedurl] = new Feed(feedurl);
+    }
 }
 
 void DataManager::loadEntry(const QString id) const
@@ -615,21 +589,141 @@ void DataManager::loadEntry(const QString id) const
 
 bool DataManager::feedExists(const QString &url)
 {
-    return m_feeds.contains(url);
+    // Try to account for some common cases where the URL is different but is
+    // actually pointing to the same data.  Currently covering:
+    // - http vs https
+    // - encoded vs non-encoded URLs
+    QString cleanUrl = QUrl(url).authority() + QUrl(url).path(QUrl::FullyDecoded);
+    for (QString listUrl : m_feedmap) {
+        if (cleanUrl == (QUrl(listUrl).authority() + QUrl(listUrl).path(QUrl::FullyDecoded))) {
+            return true;
+        }
+    }
+    return false;
 }
 
 void DataManager::updateQueueListnrs() const
 {
     QSqlQuery query;
+    query.prepare(QStringLiteral("UPDATE Queue SET listnr=:i WHERE id=:id;"));
     for (int i = 0; i < m_queuemap.count(); i++) {
-        query.prepare(QStringLiteral("UPDATE Queue SET listnr=:i WHERE id=:id;"));
         query.bindValue(QStringLiteral(":i"), i);
         query.bindValue(QStringLiteral(":id"), m_queuemap[i]);
         Database::instance().execute(query);
     }
 }
 
-bool DataManager::isFeedExists(const QString &url)
+void DataManager::bulkMarkReadByIndex(bool state, QModelIndexList list)
 {
-    return m_feeds.contains(url);
+    bulkMarkRead(state, getIdsFromModelIndexList(list));
+}
+
+void DataManager::bulkMarkRead(bool state, QStringList list)
+{
+    Database::instance().transaction();
+
+    if (state) { // Mark as read
+        // This needs special attention as the DB operations are very intensive.
+        // Reversing the loop is much faster
+        for (int i = list.count() - 1; i >= 0; i--) {
+            getEntry(list[i])->setReadInternal(state);
+        }
+        updateQueueListnrs(); // update queue after modification
+    } else { // Mark as unread
+        for (QString id : list) {
+            getEntry(id)->setReadInternal(state);
+        }
+    }
+    Database::instance().commit();
+
+    Q_EMIT bulkReadStatusActionFinished();
+
+    // if settings allow, upload these changes immediately to sync servers
+    if (state) {
+        Sync::instance().doQuickSync();
+    }
+}
+
+void DataManager::bulkMarkNewByIndex(bool state, QModelIndexList list)
+{
+    bulkMarkNew(state, getIdsFromModelIndexList(list));
+}
+
+void DataManager::bulkMarkNew(bool state, QStringList list)
+{
+    Database::instance().transaction();
+    for (QString id : list) {
+        getEntry(id)->setNewInternal(state);
+    }
+    Database::instance().commit();
+
+    Q_EMIT bulkNewStatusActionFinished();
+}
+
+void DataManager::bulkQueueStatusByIndex(bool state, QModelIndexList list)
+{
+    bulkQueueStatus(state, getIdsFromModelIndexList(list));
+}
+
+void DataManager::bulkQueueStatus(bool state, QStringList list)
+{
+    Database::instance().transaction();
+    if (state) { // i.e. add to queue
+        for (QString id : list) {
+            getEntry(id)->setQueueStatusInternal(state);
+        }
+    } else { // i.e. remove from queue
+        // This needs special attention as the DB operations are very intensive.
+        // Reversing the loop is much faster.
+        for (int i = list.count() - 1; i >= 0; i--) {
+            qCDebug(kastsDataManager) << "getting entry" << getEntry(list[i])->id();
+            getEntry(list[i])->setQueueStatusInternal(state);
+        }
+        updateQueueListnrs();
+    }
+    Database::instance().commit();
+
+    Q_EMIT bulkReadStatusActionFinished();
+    Q_EMIT bulkNewStatusActionFinished();
+}
+
+void DataManager::bulkDownloadEnclosuresByIndex(QModelIndexList list)
+{
+    bulkDownloadEnclosures(getIdsFromModelIndexList(list));
+}
+
+void DataManager::bulkDownloadEnclosures(QStringList list)
+{
+    bulkQueueStatus(true, list);
+    for (QString id : list) {
+        if (getEntry(id)->hasEnclosure()) {
+            getEntry(id)->enclosure()->download();
+        }
+    }
+}
+
+void DataManager::bulkDeleteEnclosuresByIndex(QModelIndexList list)
+{
+    bulkDeleteEnclosures(getIdsFromModelIndexList(list));
+}
+
+void DataManager::bulkDeleteEnclosures(QStringList list)
+{
+    Database::instance().transaction();
+    for (QString id : list) {
+        if (getEntry(id)->hasEnclosure()) {
+            getEntry(id)->enclosure()->deleteFile();
+        }
+    }
+    Database::instance().commit();
+}
+
+QStringList DataManager::getIdsFromModelIndexList(const QModelIndexList &list) const
+{
+    QStringList ids;
+    for (QModelIndex index : list) {
+        ids += index.data(EpisodeModel::Roles::IdRole).value<QString>();
+    }
+    qCDebug(kastsDataManager) << "Ids of selection:" << ids;
+    return ids;
 }
